<?php
/**
 * Perfect Checkout - Information controller.
 *
 * Handles the request to get store information to help with Perfect Checkout dashboard.
 *
 * @author  Sébastien Dumont
 * @package Perfect Checkout\API
 * @since   1.3.0
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * REST API Information controller class.
 *
 * @package Perfect Checkout\API
 */
class Perfect_Checkout_API_Info_Controller {

	/**
	 * Endpoint namespace.
	 *
	 * @var string
	 */
	protected $namespace = 'pc/connect';

	/**
	 * Route base.
	 *
	 * @var string
	 */
	protected $rest_base = 'info';

	/**
	 * Register routes.
	 *
	 * @access public
	 */
	public function register_routes() {
		register_rest_route(
			$this->namespace,
			'/' . $this->rest_base,
			array(
				'methods'             => WP_REST_Server::READABLE,
				'callback'            => array( $this, 'info_callback' ),
				'permission_callback' => array( $this, 'get_items_permissions_check' ),
			)
		);
	} // register_routes()

	/**
	 * Check whether a given request has permission to read info.
	 *
	 * @access public
	 *
	 * @param WP_REST_Request $request Full details about the request.
	 *
	 * @return WP_Error|boolean
	 */
	public function get_items_permissions_check( $request ) {
		if ( ! wc_rest_check_user_permissions( 'read' ) ) {
			return new WP_Error( 'perfectcheckout_rest_cannot_view', __( 'Sorry, you cannot list resources.', 'perfect-checkout' ), array( 'status' => rest_authorization_required_code() ) );
		}

		return true;
	} // END get_items_permissions_check()

	/**
	 * Returns information about the store.
	 *
	 * @access public
	 *
	 * @param WP_REST_Request $request Full details about the request.
	 *
	 * @return WP_REST_Response
	 */
	public function info_callback( $request ) {
		// Get plugins.
		if ( ! function_exists( 'get_plugins' ) ) {
			require_once ABSPATH . 'wp-admin/includes/plugin.php';
		}

		$active_plugins    = get_option( 'active_plugins' );
		$activated_plugins = array();
		$all_plugins       = get_plugins();

		// Only return active plugins.
		foreach ( $all_plugins as $plugin_file => $data ) {
			if ( in_array( $plugin_file, $active_plugins ) ) {
				$activated_plugins[ $plugin_file ] = $data;
			}
		}

		// Prepare response.
		$data = array_merge( Perfect_Checkout_Helpers::collect_additional_data(), array(
			'active_plugins' => $activated_plugins,
		) );

		// Response.
		$response = rest_ensure_response( $data );
		$response->set_status( 200 );

		return $response;
	} // END info_callback()

} // END class
