<?php
/**
 * REST API: Perfect_Checkout_API_Test_Mode_Controller class.
 *
 * @author  Sébastien Dumont
 * @package Perfect Checkout\API
 * @since   1.6.5 Introduced.
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * Controller for switching test mode on the site.
 *
 * This REST API controller handles the request to set test mode
 * via "pc/connect/test-mode" endpoint.
 *
 * @since 1.6.5 Introduced.
 */
class Perfect_Checkout_API_Test_Mode_Controller {

	/**
	 * Endpoint namespace.
	 *
	 * @var string
	 */
	protected $namespace = 'pc/connect';

	/**
	 * Route base.
	 *
	 * @var string
	 */
	protected $rest_base = 'test-mode';

	/**
	 * Register routes.
	 *
	 * @access public
	 *
	 * @ignore Function ignored when parsed into Code Reference.
	 */
	public function register_routes() {
		register_rest_route(
			$this->namespace,
			'/' . $this->rest_base,
			array(
				array(
					'methods'             => WP_REST_Server::READABLE,
					'callback'            => array( $this, 'check_status' ),
					'permission_callback' => array( $this, 'get_items_permissions_check' ),
				),
				array(
					'methods'             => WP_REST_Server::EDITABLE,
					'callback'            => array( $this, 'switch_test' ),
					'permission_callback' => array( $this, 'update_items_permissions_check' ),
					'args'                => $this->get_collection_params()
				),
			),
		);
	} // register_routes()

	/**
	 * Need write/create permission.
	 *
	 * @access public
	 *
	 * @param WP_REST_Request $request Full details about the request.
	 *
	 * @return WP_Error|boolean
	 */
	public function update_items_permissions_check( $request ) {
		if ( ! wc_rest_check_user_permissions( 'create' ) ) {
			return new WP_Error( 'perfectcheckout_rest_cannot_create', __( 'Sorry, you are not allowed to create resources.', 'perfect-checkout' ), array( 'status' => rest_authorization_required_code() ) );
		}

		return true;
	} // END update_items_permissions_check()

	/**
	 * Check whether a given request has permission to read info.
	 *
	 * @access public
	 *
	 * @param WP_REST_Request $request Full details about the request.
	 *
	 * @return WP_Error|boolean
	 */
	public function get_items_permissions_check( $request ) {
		if ( ! wc_rest_check_user_permissions( 'read' ) ) {
			return new WP_Error( 'perfectcheckout_rest_cannot_view', __( 'Sorry, you cannot list resources.', 'perfect-checkout' ), array( 'status' => rest_authorization_required_code() ) );
		}

		return true;
	} // END get_items_permissions_check()

	/**
	 * Switches the PC test mode.
	 *
	 * @access public
	 *
	 * @param WP_REST_Request $request Full details about the request.
	 *
	 * @return WP_REST_Response
	 */
	public function switch_test( $request ) {
		$params = $request->get_params();

		if ( ! empty( $params['status'] ) ) {
			if ( $params['status'] === 'off' ) {
				delete_option( 'perfectcheckout_checkout_test_mode' );
			} else {
				update_option( 'perfectcheckout_checkout_test_mode', trim( $params['status'] ) );
			}
		}

		$results = sprintf( __( "Perfect Checkout test mode has been set to '%s'", 'perfect-checkout' ), get_option( 'perfectcheckout_checkout_test_mode', 'off' ) );

		return new WP_REST_Response( $results );
	} // END switch_test()

	/**
	 * Checks the status if test mode is enabled.
	 *
	 * @access public
	 *
	 * @param WP_REST_Request $request Full details about the request.
	 *
	 * @return WP_REST_Response
	 */
	public function check_status( $request ) {
		$status = get_option( 'perfectcheckout_checkout_test_mode', 'off' );

		if ( 'off' === $status ) {
			$results = __( "Test mode is 'OFF'", 'perfect-checkout' );
		} else {
			$results = sprintf( __( "Perfect Checkout test mode is set to '%s'", 'perfect-checkout' ), $status );
		}

		return new WP_REST_Response( $results );
	} // END check_status()

	/**
	 * Get the query params.
	 *
	 * @access public
	 *
	 * @return array $params
	 */
	public function get_collection_params() { // phpcs:ignore Generic.CodeAnalysis.UselessOverridingMethod.Found
		return array(
			'status' => array(
				'description'       => __( 'Determines if Perfect Checkout is in test mode for the site.', 'perfect-checkout' ),
				'type'              => 'string',
				'enum'              => array( 'off', 'demo', 'sandbox' ),
				'required'          => true,
				'sanitize_callback' => 'sanitize_key',
				'validate_callback' => 'rest_validate_request_arg',
			),
		);
	} // END get_collection_params()

} // END class